# run with "python3 src/blueair_api/stub.py"
import logging
import asyncio
from threading import Event

from getpass import getpass
from pathlib import Path
import sys

# import blueair_api

path_root = Path(__file__).parents[2]
sys.path.append(str(path_root))
from src.blueair_api import get_devices, get_aws_devices


logger = logging.getLogger("src.blueair_api")
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)


async def testing():
    username = input("Username: ")
    password = getpass()
    try:
        api, devices = await get_aws_devices(username=username, password=password)
        for device in devices:
            await device.refresh()
            await device.set_child_lock(True)
            logger.debug(device)
    finally:
        if api:
            await api.cleanup_client_session()
    try:
        api, devices = await get_devices(username=username, password=password)
        for device in devices:
            await device.refresh()
            logger.debug(device)
    finally:
        if api:
            await api.cleanup_client_session()


asyncio.run(testing())



import dataclasses
import logging

from .callbacks import CallbacksMixin
from .http_blueair import HttpBlueair
from .util import transform_data_points, safely_get_json_value

_LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass(slots=True)
class Device(CallbacksMixin):
    @classmethod
    async def create_device(cls, api, uuid, name, mac, refresh=False):
        _LOGGER.debug("UUID:"+uuid)
        info = await api.get_info(uuid)
        device = Device(
            api=api,
            uuid=uuid,
            name=name,
            mac=mac,
            timezone=info["timezone"],
            compatibility=info["compatibility"],
            model=info["model"],
            firmware=info["firmware"],
            mcu_firmware=info["mcuFirmware"],
            wlan_driver=info["wlanDriver"],
            room_location=info["roomLocation"]
        )
        if refresh:
            await device.refresh()
        _LOGGER.debug(f"create_device blueair device: {device}")
        return device

    api: HttpBlueair
    uuid: str | None = None
    name: str | None = None
    timezone: str | None = None
    compatibility: str | None = None
    model: str | None = None
    mac: str | None = None
    firmware: str | None = None
    mcu_firmware: str | None = None
    wlan_driver: str | None = None
    room_location: str | None = None

    brightness: int | None = None
    child_lock: bool | None = None
    night_mode: bool | None = None
    fan_speed: int | None = None
    fan_mode: str | None = None
    filter_expired: bool | None = None
    wifi_working: bool | None = None

    pm1: int | None = None
    pm10: int | None = None
    pm25: int | None = None
    voc: int | None = None
    co2: int | None = None
    temperature: float | None = None
    humidity: float | None = None
    all_pollution: float | None = None

    async def refresh(self):
        _LOGGER.debug("Requesting current attributes...")
        attributes = await self.api.get_attributes(self.uuid)
        _LOGGER.debug(f"result: {attributes}")
        if "brightness" in attributes:
            self.brightness = int(attributes["brightness"])
        else:
            self.brightness = 0
        if "child_lock" in attributes:
            self.child_lock = attributes["child_lock"] == "1"
        if "night_mode" in attributes:
            self.night_mode = bool(attributes["night_mode"])
        self.fan_speed = int(attributes["fan_speed"])
        if "filter_status" in attributes:
            self.filter_expired = attributes["filter_status"] != "OK"
        self.fan_mode = attributes["mode"]
        if "wifi_status" in attributes:
            self.wifi_working = attributes["wifi_status"] == "1"
        else:
            self.wifi_working = False
        if self.compatibility != "sense+":
            for data_point in transform_data_points(await self.api.get_data_points_since(self.uuid)):
                _LOGGER.debug(data_point)
                self.pm25 = safely_get_json_value(data_point, "pm25", int)
                self.pm10 = safely_get_json_value(data_point, "pm10", int)
                self.pm1 = safely_get_json_value(data_point, "pm1", int)
                self.voc = safely_get_json_value(data_point, "voc", int)
                self.co2 = safely_get_json_value(data_point, "co2", int)
                self.temperature = safely_get_json_value(data_point, "temperature", int)
                self.humidity = safely_get_json_value(data_point, "humidity", int)
                self.all_pollution = safely_get_json_value(data_point, "all_pollution", int)
        _LOGGER.debug(f"refreshed blueair device: {self}")
        self.publish_updates()

    async def set_fan_speed(self, new_speed):
        self.fan_speed = new_speed
        await self.api.set_fan_speed(self.uuid, new_speed)
        self.publish_updates()



import logging

from aiohttp import ClientSession

from .http_blueair import HttpBlueair
from .http_aws_blueair import HttpAwsBlueair
from .device import Device
from .device_aws import DeviceAws
from typing import Optional

_LOGGER = logging.getLogger(__name__)


async def get_devices(
    username: str,
    password: str,
    home_host: str | None = None,
    auth_token: str | None = None,
    client_session: ClientSession | None = None,
) -> tuple[HttpBlueair, list[Device]]:
    api = HttpBlueair(
        client_session=client_session,
        username=username,
        password=password,
        home_host=home_host,
        auth_token=auth_token,
    )
    api_devices = await api.get_devices()
    devices = []
    for api_device in api_devices:
        devices.append(await Device.create_device(
            api=api,
            uuid=api_device["uuid"],
            name=api_device["name"],
            mac=api_device["mac"]
        ))
    return (
        api,
        devices,
    )


async def get_aws_devices(
    username: str,
    password: str,
    region: str = "us",
    client_session: ClientSession | None = None,
) -> tuple[HttpAwsBlueair, list[Device]]:
    api = HttpAwsBlueair(
        username=username,
        password=password,
        region=region,
        client_session=client_session,
    )
    api_devices = await api.devices()
    devices = []
    for api_device in api_devices:
        _LOGGER.debug("api_device: %s", api_device)
        devices.append(await DeviceAws.create_device(
            api=api,
            uuid=api_device["uuid"],
            name=api_device["name"],
            mac=api_device["mac"],
            type_name=api_device["type"]
        ))
    return (
        api,
        devices
    )



from collections.abc import Mapping
from typing import TypedDict

SENSITIVE_FIELD_NAMES = [
    "username",
    "password",
]

# The BlueAir API uses a fixed API key.
API_KEY = "eyJhbGciOiJIUzI1NiJ9.eyJncmFudGVlIjoiYmx1ZWFpciIsImlhdCI6MTQ1MzEyNTYzMiwidmFsaWRpdHkiOi0xLCJqdGkiOiJkNmY3OGE0Yi1iMWNkLTRkZDgtOTA2Yi1kN2JkNzM0MTQ2NzQiLCJwZXJtaXNzaW9ucyI6WyJhbGwiXSwicXVvdGEiOi0xLCJyYXRlTGltaXQiOi0xfQ.CJsfWVzFKKDDA6rWdh-hjVVVE9S3d6Hu9BzXG9htWFw"  # noqa: E501

AWS_APIKEYS = {
    "us": {
        "gigyaRegion": "us1",
        "restApiId": "on1keymlmh",
        "awsRegion": "us-east-2",
        "apiKey": "3_-xUbbrIY8QCbHDWQs1tLXE-CZBQ50SGElcOY5hF1euE11wCoIlNbjMGAFQ6UwhMY",
    },
    "eu": {
        "gigyaRegion": "eu1",
        "restApiId": "hkgmr8v960",
        "awsRegion": "eu-west-1",
        "apiKey": "3_qRseYzrUJl1VyxvSJANalu_kNgQ83swB1B9uzgms58--5w1ClVNmrFdsDnWVQQCl",
    },
}


class MeasurementBundle(TypedDict):
    sensors: list[str]
    datapoints: list[list[int | float]]


MeasurementList = list[Mapping[str, int | float]]



import logging

_LOGGER = logging.getLogger(__name__)


class CallbacksMixin:
    __slots__ = ['_callbacks']

    def _setup_callbacks(self):
        self._callbacks = set()

    def register_callback(self, callback) -> None:
        if not hasattr(self, "_callbacks"):
            self._setup_callbacks()
        self._callbacks.add(callback)

    def remove_callback(self, callback) -> None:
        if not hasattr(self, "_callbacks"):
            self._setup_callbacks()
        self._callbacks.discard(callback)

    def publish_updates(self) -> None:
        if not hasattr(self, "_callbacks"):
            self._setup_callbacks()
        _LOGGER.debug(f"{id(self)} publishing updates")
        for callback in self._callbacks:
            callback()



import dataclasses
import logging
from json import dumps

from .callbacks import CallbacksMixin
from .http_aws_blueair import HttpAwsBlueair
from .model_enum import ModelEnum
from . import intermediate_representation_aws as ir

_LOGGER = logging.getLogger(__name__)

type AttributeType[T] = T | None

@dataclasses.dataclass(slots=True)
class DeviceAws(CallbacksMixin):
    @classmethod
    async def create_device(cls, api, uuid, name, mac, type_name, refresh=False):
        _LOGGER.debug("UUID:"+uuid)
        device_aws = DeviceAws(
            api=api,
            uuid=uuid,
            name_api=name,
            mac=mac,
            type_name=type_name,
        )
        if refresh:
            await device_aws.refresh()
        _LOGGER.debug(f"create_device blueair device_aws: {device_aws}")
        return device_aws

    api: HttpAwsBlueair
    uuid : str | None = None
    name : str | None = None
    name_api : str | None = None
    mac : str | None = None
    type_name : str | None = None

    # Attributes are defined below.
    # We mandate that unittests shall test all fields of AttributeType.
    sku : AttributeType[str] = None
    firmware : AttributeType[str] = None
    mcu_firmware : AttributeType[str] = None
    serial_number : AttributeType[str] = None

    brightness : AttributeType[int] = None
    child_lock : AttributeType[bool] = None
    fan_speed : AttributeType[int] = None
    fan_auto_mode : AttributeType[bool] = None
    standby : AttributeType[bool] = None
    night_mode : AttributeType[bool] = None
    germ_shield : AttributeType[bool] = None

    pm1 : AttributeType[int] = None
    pm2_5 : AttributeType[int] = None
    pm10 : AttributeType[int] = None
    tVOC : AttributeType[int] = None
    temperature : AttributeType[int] = None
    humidity : AttributeType[int] = None
    filter_usage_percentage : AttributeType[int] = None
    wifi_working : AttributeType[bool] = None

    wick_usage_percentage : AttributeType[int] = None
    wick_dry_mode : AttributeType[bool] = None
    water_shortage : AttributeType[bool] = None
    auto_regulated_humidity : AttributeType[int] = None

    async def refresh(self):
        _LOGGER.debug(f"refreshing blueair device aws: {self}")
        info = await self.api.device_info(self.name_api, self.uuid)
        _LOGGER.debug(dumps(info, indent=2))

        # ir.parse_json(ir.Attribute, ir.query_json(info, "configuration.da"))
        ds = ir.parse_json(ir.Sensor, ir.query_json(info, "configuration.ds"))
        dc = ir.parse_json(ir.Control, ir.query_json(info, "configuration.dc"))

        sensor_data = ir.SensorPack(info["sensordata"]).to_latest_value()

        def sensor_data_safe_get(key):
            return sensor_data.get(key) if key in ds else NotImplemented

        self.pm1 = sensor_data_safe_get("pm1")
        self.pm2_5 = sensor_data_safe_get("pm2_5")
        self.pm10 = sensor_data_safe_get("pm10")
        self.tVOC = sensor_data_safe_get("tVOC")
        self.temperature = sensor_data_safe_get("t")
        self.humidity = sensor_data_safe_get("h")

        def info_safe_get(path):
            # directly reads for the schema. If the schema field is
            # undefined, it is NotImplemented, not merely unavailable.
            value = ir.query_json(info, path)
            if value is None:
                return NotImplemented
            return value

        self.name = info_safe_get("configuration.di.name")
        self.firmware = info_safe_get("configuration.di.cfv")
        self.mcu_firmware = info_safe_get("configuration.di.mfv")
        self.serial_number = info_safe_get("configuration.di.ds")
        self.sku = info_safe_get("configuration.di.sku")

        states = ir.SensorPack(info["states"]).to_latest_value()

        def states_safe_get(key):
            return states.get(key) if key in dc else NotImplemented

        # "online" is not defined in the schema.
        self.wifi_working = states.get("online")

        self.standby = states_safe_get("standby")
        self.night_mode = states_safe_get("nightmode")
        self.germ_shield = states_safe_get("germshield")
        self.brightness = states_safe_get("brightness")
        self.child_lock = states_safe_get("childlock")
        self.fan_speed = states_safe_get("fanspeed")
        self.fan_auto_mode = states_safe_get("automode")
        self.filter_usage_percentage = states_safe_get("filterusage")
        self.wick_usage_percentage = states_safe_get("wickusage")
        self.wick_dry_mode = states_safe_get("wickdrys")
        self.auto_regulated_humidity = states_safe_get("autorh")
        self.water_shortage = states_safe_get("wshortage")

        self.publish_updates()
        _LOGGER.debug(f"refreshed blueair device aws: {self}")

    async def set_brightness(self, value: int):
        self.brightness = value
        await self.api.set_device_info(self.uuid, "brightness", "v", value)
        self.publish_updates()

    async def set_fan_speed(self, value: int):
        self.fan_speed = value
        await self.api.set_device_info(self.uuid, "fanspeed", "v", value)
        self.publish_updates()

    async def set_standby(self, value: bool):
        self.standby = value
        await self.api.set_device_info(self.uuid, "standby", "vb", value)
        self.publish_updates()

    # FIXME: avoid state translation at the API level and depreate running.
    # replace with standby which is standard across aws devices.
    @property
    def running(self) -> AttributeType[bool]:
        if self.standby is None or self.standby is NotImplemented:
            return self.standby
        return not self.standby

    async def set_running(self, running: bool):
        await self.set_standby(not running)

    async def set_fan_auto_mode(self, fan_auto_mode: bool):
        self.fan_auto_mode = fan_auto_mode
        await self.api.set_device_info(self.uuid, "automode", "vb", fan_auto_mode)
        self.publish_updates()

    async def set_auto_regulated_humidity(self, value: int):
        self.auto_regulated_humidity = value
        await self.api.set_device_info(self.uuid, "autorh", "v", value)
        self.publish_updates()

    async def set_child_lock(self, child_lock: bool):
        self.child_lock = child_lock
        await self.api.set_device_info(self.uuid, "childlock", "vb", child_lock)
        self.publish_updates()

    async def set_night_mode(self, night_mode: bool):
        self.night_mode = night_mode
        await self.api.set_device_info(self.uuid, "nightmode", "vb", night_mode)
        self.publish_updates()

    async def set_wick_dry_mode(self, value: bool):
        self.wick_dry_mode = value
        await self.api.set_device_info(self.uuid, "wickdrys", "vb", value)
        self.publish_updates()

    @property
    def model(self) -> ModelEnum:
        if self.sku == "111633":
            return ModelEnum.HUMIDIFIER_H35I
        if self.sku == "105820":
            return ModelEnum.PROTECT_7440I
        if self.sku == "105826":
            return ModelEnum.PROTECT_7470I
        if self.sku == "110059":
            return ModelEnum.MAX_211I
        if self.sku == "110092":
            return ModelEnum.MAX_311I
        if self.sku == "110057":
            return ModelEnum.MAX_411I
        if self.sku == "112124":
            return ModelEnum.T10I
        return ModelEnum.UNKNOWN




from enum import Enum, StrEnum


class FeatureEnum(StrEnum):
    TEMPERATURE = "Temperature"
    HUMIDITY = "Humidity"
    VOC = "VOC"
    PM1 = "PM1"
    PM10 = "PM10"
    PM25 = "PM25"
    WATER_SHORTAGE = "Water Shortage"
    FILTER_EXPIRED = "Filter Expired"
    CHILD_LOCK = "Child Lock"


class ModelEnum(Enum):
    def __new__(cls, *args, **kwds):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self,
                 name,
                 supported_features):
        self.model_name = name
        self.supported_features = supported_features

    def supports_feature(self, supported_features) -> bool:
        return supported_features in self.supported_features

    UNKNOWN = "Unknown", [
        FeatureEnum.TEMPERATURE,
        FeatureEnum.HUMIDITY,
        FeatureEnum.WATER_SHORTAGE,
        FeatureEnum.VOC,
        FeatureEnum.PM1,
        FeatureEnum.PM10,
        FeatureEnum.PM25,
        FeatureEnum.FILTER_EXPIRED,
        FeatureEnum.CHILD_LOCK,
    ]
    HUMIDIFIER_H35I = "Blueair Humidifier H35i", [
        FeatureEnum.TEMPERATURE,
        FeatureEnum.HUMIDITY,
        FeatureEnum.WATER_SHORTAGE,
    ]
    PROTECT_7440I = "Blueair Protect 7440i", [
        FeatureEnum.TEMPERATURE,
        FeatureEnum.HUMIDITY,
        FeatureEnum.VOC,
        FeatureEnum.PM1,
        FeatureEnum.PM10,
        FeatureEnum.PM25,
        FeatureEnum.FILTER_EXPIRED,
        FeatureEnum.CHILD_LOCK,
    ]
    PROTECT_7470I = "Blueair Protect 7470i", [
        FeatureEnum.TEMPERATURE,
        FeatureEnum.HUMIDITY,
        FeatureEnum.VOC,
        FeatureEnum.PM1,
        FeatureEnum.PM10,
        FeatureEnum.PM25,
        FeatureEnum.FILTER_EXPIRED,
        FeatureEnum.CHILD_LOCK,
    ]
    MAX_211I = "Blueair Blue Pure 211i Max", [
        FeatureEnum.PM1,
        FeatureEnum.PM10,
        FeatureEnum.PM25,
        FeatureEnum.FILTER_EXPIRED,
        FeatureEnum.CHILD_LOCK,
    ]
    MAX_311I = "Blueair Blue Pure 311i Max", [
        FeatureEnum.PM25,
        FeatureEnum.FILTER_EXPIRED,
        FeatureEnum.CHILD_LOCK,
    ]
    MAX_411I = "Blueair Blue Pure 411i Max", [
        FeatureEnum.PM25,
        FeatureEnum.FILTER_EXPIRED,
        FeatureEnum.CHILD_LOCK,
    ]
    T10I = "T10i ComfortPure 3-in-1 Filter/Heater/Fan", [
        FeatureEnum.TEMPERATURE,
        FeatureEnum.HUMIDITY,
        FeatureEnum.PM25,
        FeatureEnum.FILTER_EXPIRED,
    ]



from typing import Any
import logging

from aiohttp import ClientSession, ClientResponse
import base64

from .util_http import request_with_logging
from .const import API_KEY
from .errors import LoginError
from typing import Optional

_LOGGER = logging.getLogger(__name__)


class HttpBlueair:
    def __init__(
        self,
        username: str,
        password: str,
        home_host: str | None = None,
        auth_token: str | None = None,
        client_session: ClientSession | None = None,
    ):
        self.username = username
        self.password = password
        self.home_host = home_host
        self.auth_token = auth_token

        if client_session is None:
            self.api_session = ClientSession(raise_for_status=False)
        else:
            self.api_session = client_session

    async def cleanup_client_session(self):
        await self.api_session.close()

    async def get_home_host(self):
        if self.home_host is None:
            self.home_host = await self._get_home_host()
        return self.home_host

    async def get_auth_token(self):
        if self.auth_token is None:
            self.auth_token = await self._get_auth_token()
        return self.auth_token

    @request_with_logging
    async def _get_request_with_logging_and_errors_raised(
        self, url: str, headers: dict | None = None
    ) -> ClientResponse:
        return await self.api_session.get(url=url, headers=headers)

    @request_with_logging
    async def _post_request_with_logging_and_errors_raised(
        self, url: str, json_body: dict, headers: dict | None = None
    ) -> ClientResponse:
        return await self.api_session.post(url=url, json=json_body, headers=headers)

    async def _get_home_host(self) -> str:
        """
        Retrieve the home host for the current username.

        The home host is the server that is used to interact with the Blueair
        device. It can be stored and reused to avoid requesting it again when
        reinitializing the class at a later time.
        """
        url = f"https://api.blueair.io/v2/user/{self.username}/homehost/"
        headers = {"X-API-KEY-TOKEN": API_KEY}
        response: ClientResponse = (
            await self._get_request_with_logging_and_errors_raised(
                url=url, headers=headers
            )
        )
        if response.status == 404:
            raise LoginError("invalid username")
        return (await response.text()).replace('"', "")

    async def _get_auth_token(self) -> str:
        """
        Authenticate the user and retrieve the authentication token.

        The authentication token can be reused to prevent an additional network
        request when initializing the client.
        """
        url = f"https://{await self.get_home_host()}/v2/user/{self.username}/login/"
        headers = {
            "X-API-KEY-TOKEN": API_KEY,
            "Authorization": "Basic "
            + base64.b64encode((self.username + ":" + self.password).encode()).decode(),
        }
        response: ClientResponse = (
            await self._get_request_with_logging_and_errors_raised(
                url=url, headers=headers
            )
        )
        result = await response.text()
        if response.status == 404:
            raise LoginError("invalid username")
        if result == "true":
            return response.headers["X-AUTH-TOKEN"]
        else:
            raise LoginError("invalid password")

    async def get_devices(self) -> list[dict[str, Any]]:
        """
        Fetch a list of devices.

        Returns a list of dictionaries. Each dictionary will have a UUID key
        (the device identifier), a user ID, MAC address, and device name.

        Example response:

        [{"uuid":"1234567890ABCDEF","userId":12345,"mac":"1234567890AB","name":"My Blueair Device"}]
        """
        url = f"https://{await self.get_home_host()}/v2/owner/{self.username}/device/"
        headers = {
            "X-API-KEY-TOKEN": API_KEY,
            "X-AUTH-TOKEN": await self.get_auth_token(),
        }
        response: ClientResponse = (
            await self._get_request_with_logging_and_errors_raised(
                url=url, headers=headers
            )
        )
        return await response.json()

    # Note: refreshes every 5 minutes
    async def get_attributes(self, device_uuid: str) -> dict[str, Any]:
        """
        Fetch a list of attributes for the provided device ID.

        The return value is a dictionary containing key-value pairs for any
        available attributes.

        Note: the data for this API call is only updated once every 5 minutes.
        Calling it more often will return the same respone from the server and
        should be avoided to limit server load.
        """
        url = (
            f"https://{await self.get_home_host()}/v2/device/{device_uuid}/attributes/"
        )
        headers = {
            "X-API-KEY-TOKEN": API_KEY,
            "X-AUTH-TOKEN": await self.get_auth_token(),
        }
        response: ClientResponse = (
            await self._get_request_with_logging_and_errors_raised(
                url=url, headers=headers
            )
        )
        raw_attributes = await response.json()
        attributes = {}
        for item in raw_attributes:
            attributes[item["name"]] = item["currentValue"]

        return attributes

    # Note: refreshes every 5 minutes, timestamps are in seconds
    async def get_info(self, device_uuid: str) -> dict[str, Any]:
        """
        Fetch device information for the provided device ID.

        The return value is a dictionary containing key-value pairs for the
        available device information.

        Note: the data for this API call is only updated once every 5 minutes.
        Calling it more often will return the same respone from the server and
        should be avoided to limit server load.
        """
        url = f"https://{await self.get_home_host()}/v2/device/{device_uuid}/info/"
        headers = {
            "X-API-KEY-TOKEN": API_KEY,
            "X-AUTH-TOKEN": await self.get_auth_token(),
        }
        response: ClientResponse = (
            await self._get_request_with_logging_and_errors_raised(
                url=url, headers=headers
            )
        )
        return await response.json()

    # Note: refreshes every 5 minutes, timestamps are in seconds
    async def get_current_data_point(self, device_uuid: str) -> dict[str, Any]:
        """
        Fetch device information for the provided device ID.

        The return value is a dictionary containing key-value pairs for the
        available device information.

        Note: the data for this API call is only updated once every 5 minutes.
        Calling it more often will return the same response from the server and
        should be avoided to limit server load.
        """
        url = f"https://{await self.get_home_host()}/v2/device/{device_uuid}/datapoint/0/last/0/"
        headers = {
            "X-API-KEY-TOKEN": API_KEY,
            "X-AUTH-TOKEN": await self.get_auth_token(),
        }
        response: ClientResponse = (
            await self._get_request_with_logging_and_errors_raised(
                url=url, headers=headers
            )
        )
        return await response.json()

    async def get_data_points_since(self, device_uuid: str, seconds_ago: int = 0, sample_period: int = 300) -> dict[str, Any]:
        """
        Fetch the list of data points between a relative timestamp (in seconds) and the current time.

        An optional sample period can be provided to group data points
        together. The minimum sample period size is 300 (5 minutes).

        Note: the data for the most recent data point is only updated once
        every 5 minutes.  Calling it more often will return the same response
        from the server and should be avoided to limit server load.
        """
        url = f"https://{await self.get_home_host()}/v2/device/{device_uuid}/datapoint/{seconds_ago}/last/{sample_period}/"
        headers = {
            "X-API-KEY-TOKEN": API_KEY,
            "X-AUTH-TOKEN": await self.get_auth_token(),
        }
        response: ClientResponse = (
            await self._get_request_with_logging_and_errors_raised(
                url=url, headers=headers
            )
        )
        return await response.json()

    async def set_fan_speed(self, device_uuid, new_speed: str):
        """
        Set the fan speed per @spikeyGG comment at https://community.home-assistant.io/t/blueair-purifier-addon/154456/14
        """
        if new_speed == "auto":
            new_name = "mode"
        elif new_speed in ["0", "1", "2", "3"]:
            new_name = "fan_speed"
        else:
            raise Exception("Speed not supported")
        url = f"https://{await self.get_home_host()}/v2/device/{device_uuid}/attribute/fanspeed/"
        headers = {
            "X-API-KEY-TOKEN": API_KEY,
            "X-AUTH-TOKEN": await self.get_auth_token(),
        }
        json_body = {
            "currentValue": new_speed,
            "scope": "device",
            "defaultValue": new_speed,
            "name": new_name,
            "uuid": str(device_uuid),
        }
        response: ClientResponse = (
            await self._post_request_with_logging_and_errors_raised(
                url=url, json_body=json_body, headers=headers
            )
        )
        return await response.json()



from typing import Any
import functools
import logging

from aiohttp import ClientSession, ClientResponse, FormData
from .const import AWS_APIKEYS

from .util_http import request_with_logging
from .errors import SessionError, LoginError

_LOGGER = logging.getLogger(__name__)


def request_with_active_session(func):
    @functools.wraps(func)
    async def request_with_active_session_wrapper(*args, **kwargs) -> ClientResponse:
        _LOGGER.debug("session")
        try:
            return await func(*args, **kwargs)
        except SessionError:
            _LOGGER.debug("got invalid session, attempting to repair and resend")
            self = args[0]
            self.session_token = None
            self.session_secret = None
            self.access_token = None
            self.jwt = None
            response = await func(*args, **kwargs)
            return response

    return request_with_active_session_wrapper


def request_with_errors(func):
    @functools.wraps(func)
    async def request_with_errors_wrapper(*args, **kwargs) -> ClientResponse:
        _LOGGER.debug("checking for errors")
        response: ClientResponse = await func(*args, **kwargs)
        status_code = response.status
        try:
            response_json = await response.json(content_type=None)
            if "statusCode" in response_json:
                _LOGGER.debug("response json found, checking status code from response")
                status_code = response_json["statusCode"]
        except Exception as e:
            _LOGGER.error(f"Error parsing response for errors {e}")
            raise e
        if status_code == 200:
            _LOGGER.debug("response 200")
            return response
        if 400 <= status_code <= 500:
            _LOGGER.debug("auth error")
            url = kwargs["url"]
            response_text = await response.text()
            if "accounts.login" in url:
                _LOGGER.debug("login error")
                raise LoginError(response_text)
            else:
                _LOGGER.debug("session error")
                raise SessionError(response_text)
        raise ValueError(f"unknown status code {status_code}")

    return request_with_errors_wrapper


class HttpAwsBlueair:
    def __init__(
        self,
        username: str,
        password: str,
        region: str = "us",
        client_session: ClientSession | None = None,
    ):
        self.username = username
        self.password = password
        self.region = region

        self.session_token = None
        self.session_secret = None

        self.access_token = None

        self.jwt = None

        if client_session is None:
            self.api_session = ClientSession(raise_for_status=False)
        else:
            self.api_session = client_session

    async def cleanup_client_session(self):
        await self.api_session.close()

    @request_with_errors
    @request_with_logging
    async def _get_request_with_logging_and_errors_raised(
        self, url: str, headers: dict | None = None
    ) -> ClientResponse:
        return await self.api_session.get(url=url, headers=headers)

    @request_with_errors
    @request_with_logging
    async def _post_request_with_logging_and_errors_raised(
        self,
        url: str,
        json_body: dict | None = None,
        form_data: FormData | None = None,
        headers: dict | None = None,
    ) -> ClientResponse:
        return await self.api_session.post(
            url=url, data=form_data, json=json_body, headers=headers
        )

    async def refresh_session(self) -> None:
        _LOGGER.debug("refresh_session")
        url = f"https://accounts.{AWS_APIKEYS[self.region]['gigyaRegion']}.gigya.com/accounts.login"
        form_data = FormData()
        form_data.add_field("apikey", AWS_APIKEYS[self.region]["apiKey"])
        form_data.add_field("loginID", self.username)
        form_data.add_field("password", self.password)
        form_data.add_field("targetEnv", "mobile")
        response: ClientResponse = (
            await self._post_request_with_logging_and_errors_raised(
                url=url, form_data=form_data
            )
        )
        response_json = await response.json(content_type="text/javascript")
        self.session_token = response_json["sessionInfo"]["sessionToken"]
        self.session_secret = response_json["sessionInfo"]["sessionSecret"]

    async def refresh_jwt(self) -> None:
        _LOGGER.debug("refresh_jwt")
        if self.session_token is None or self.session_secret is None:
            await self.refresh_session()
        url = f"https://accounts.{AWS_APIKEYS[self.region]['gigyaRegion']}.gigya.com/accounts.getJWT"
        form_data = FormData()
        form_data.add_field("oauth_token", self.session_token)
        form_data.add_field("secret", self.session_secret)
        form_data.add_field("targetEnv", "mobile")
        response: ClientResponse = (
            await self._post_request_with_logging_and_errors_raised(
                url=url, form_data=form_data
            )
        )
        response_json = await response.json(content_type="text/javascript")
        self.jwt = response_json["id_token"]

    async def refresh_access_token(self) -> None:
        _LOGGER.debug("refresh_access_token")
        if self.jwt is None:
            await self.refresh_jwt()
        url = f"https://{AWS_APIKEYS[self.region]['restApiId']}.execute-api.{AWS_APIKEYS[self.region]['awsRegion']}.amazonaws.com/prod/c/login"
        headers = {"idtoken": self.jwt}
        response: ClientResponse = (
            await self._post_request_with_logging_and_errors_raised(
                url=url, headers=headers
            )
        )
        response_json = await response.json()
        self.access_token = response_json["access_token"]

    async def get_access_token(self) -> str:
        _LOGGER.debug("get_access_token")
        if self.access_token is None:
            await self.refresh_access_token()
        assert self.access_token is not None
        return self.access_token

    @request_with_active_session
    async def devices(self) -> dict[str, Any]:
        _LOGGER.debug("devices")
        url = f"https://{AWS_APIKEYS[self.region]['restApiId']}.execute-api.{AWS_APIKEYS[self.region]['awsRegion']}.amazonaws.com/prod/c/registered-devices"
        headers = {
            "Authorization": f"Bearer {await self.get_access_token()}",
        }
        response: ClientResponse = (
            await self._get_request_with_logging_and_errors_raised(
                url=url, headers=headers
            )
        )
        response_json = await response.json()
        return response_json["devices"]

    @request_with_active_session
    async def device_info(self, device_name, device_uuid) -> dict[str, Any]:
        _LOGGER.debug("device_info")
        url = f"https://{AWS_APIKEYS[self.region]['restApiId']}.execute-api.{AWS_APIKEYS[self.region]['awsRegion']}.amazonaws.com/prod/c/{device_name}/r/initial"
        headers = {
            "Authorization": f"Bearer {await self.get_access_token()}",
        }
        json_body = {
            "deviceconfigquery": [
                {
                    "id": device_uuid,
                    "r": {
                        "r": [
                            "sensors",
                        ],
                    },
                },
            ],
            "includestates": True
        }
        response: ClientResponse = (
            await self._post_request_with_logging_and_errors_raised(
                url=url, headers=headers, json_body=json_body
            )
        )
        response_json = await response.json()
        return response_json["deviceInfo"][0]

    @request_with_active_session
    async def set_device_info(
        self, device_uuid, service_name, action_verb, action_value
    ) -> bool:
        _LOGGER.debug("set_device_info")
        url = f"https://{AWS_APIKEYS[self.region]['restApiId']}.execute-api.{AWS_APIKEYS[self.region]['awsRegion']}.amazonaws.com/prod/c/{device_uuid}/a/{service_name}"
        headers = {
            "Authorization": f"Bearer {await self.get_access_token()}",
        }
        json_body = {"n": service_name, action_verb: action_value}
        response: ClientResponse = (
            await self._post_request_with_logging_and_errors_raised(
                url=url, headers=headers, json_body=json_body
            )
        )
        response_text = await response.text()
        return response_text == "Success"



from typing import Any
import logging

from .const import SENSITIVE_FIELD_NAMES

_LOGGER = logging.getLogger(__name__)


def clean_dictionary_for_logging(dictionary: dict[str, Any]) -> dict[str, Any]:
    mutable_dictionary = dictionary.copy()
    for key in dictionary:
        if key.lower() in SENSITIVE_FIELD_NAMES:
            mutable_dictionary[key] = "***"
        if type(mutable_dictionary[key]) is dict:
            mutable_dictionary[key] = clean_dictionary_for_logging(
                mutable_dictionary[key].copy()
            )
        if type(mutable_dictionary[key]) is list:
            new_array = []
            for item in mutable_dictionary[key]:
                if type(item) is dict:
                    new_array.append(clean_dictionary_for_logging(item.copy()))
                else:
                    new_array.append(item)
            mutable_dictionary[key] = new_array

    return mutable_dictionary


def safely_get_json_value(json, key, callable_to_cast=None):
    value = json
    for x in key.split("."):
        if value is not None:
            try:
                value = value[x]
            except (TypeError, KeyError):
                try:
                    value = value[int(x)]
                except (TypeError, KeyError, ValueError):
                    value = None
    if callable_to_cast is not None and value is not None:
        value = callable_to_cast(value)
    return value


def transform_data_points(data):
    """Transform a measurement list response from the Blueair API to a more pythonic data structure."""
    key_mapping = {
        "time": "timestamp",
        "pm": "pm25",
        "pm1": "pm1",
        "pm10": "pm10",
        "tmp": "temperature",
        "hum": "humidity",
        "co2": "co2",
        "voc": "voc",
        "allpollu": "all_pollution",
    }

    keys = [key_mapping[key] for key in data["sensors"]]

    return [dict(zip(keys, values)) for values in data["datapoints"]]
